/**
 * OW2 Util
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: BasicPool.java 5226 2009-11-08 18:40:34Z gaellalire $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.pool.impl.enhanced.impl.basic;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.Executor;
import java.util.concurrent.locks.Lock;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.pool.impl.enhanced.api.IWaitControl;
import org.ow2.util.pool.impl.enhanced.api.IllegalTimeoutException;
import org.ow2.util.pool.impl.enhanced.api.NotABusyPoolItemException;
import org.ow2.util.pool.impl.enhanced.api.TimeoutPoolException;
import org.ow2.util.pool.impl.enhanced.api.WaiterInterruptedException;
import org.ow2.util.pool.impl.enhanced.api.basic.CreatePoolItemException;
import org.ow2.util.pool.impl.enhanced.api.basic.IBasicPool;
import org.ow2.util.pool.impl.enhanced.api.basic.IPoolItemFactory;
import org.ow2.util.pool.impl.enhanced.api.basic.IPoolItemRemoveListener;
import org.ow2.util.pool.impl.enhanced.api.basic.NoBusyPoolItemException;
import org.ow2.util.pool.impl.enhanced.api.basic.PoolFactoryBroken;
import org.ow2.util.pool.impl.enhanced.api.basic.accessmanager.IAccessManager;
import org.ow2.util.pool.impl.enhanced.impl.PoolError;
import org.ow2.util.pool.impl.enhanced.impl.util.LockFactory;
import org.ow2.util.pool.impl.enhanced.impl.waitcontrol.InfiniteWaitControl;
import org.ow2.util.pool.impl.enhanced.internal.actionscheduler.ActionSchedulerFactory;
import org.ow2.util.pool.impl.enhanced.internal.actionscheduler.IAction;
import org.ow2.util.pool.impl.enhanced.internal.actionscheduler.IActionScheduler;
import org.ow2.util.pool.impl.enhanced.internal.lock.api.ISignalClearableCondition;
import org.ow2.util.pool.impl.enhanced.internal.lock.impl.SignalClearableConditionProxy;

/**
 * Basic pool implementation. You can dynamically change pool size calling
 * setExpectedSize method.
 * @author Gael Lalire
 * @param <E> pool item type
 */
public class BasicPool<E> implements IBasicPool<E> {

    private static final Log LOG = LogFactory.getLog(BasicPool.class);

    private Lock lock;

    private ISignalClearableCondition signalClearableCondition;

    private int usedPoolItem;

    private List<E> availablePoolItemList;

    private List<E> unmodifiableAvailablePoolItemList;

    private IPoolItemRemoveListener<? super E> poolItemRemoveCB;

    private IPoolItemFactory<? extends E> poolItemFactory;

    private int expectedSize;

    private IAccessManager<? super E> accessManager;

    private Object sizeMutex;

    /**
     * To modify or read this field you must gain the sizeMutex.
     */
    private int brokenScheduledCreateItemCount;

    private int scheduledCreateItemCount;

    private IActionScheduler createItemActionScheduler;

    private IActionScheduler removeItemActionScheduler;

    private int delayedCount;

    /**
     * To modify this field you must gain first the sizeMutex and second the lock.
     * To read it you can gain sizeMutex or lock.
     */
    private FactoryState factoryState;

    /**
     * To modify this field you must gain first the sizeMutex and second the lock.
     * To read it you can gain sizeMutex or lock.
     */
    private long temporaryBrokenFactoryEndTime;

    private boolean sizeLockedToZero = false;

    public BasicPool(final IPoolItemFactory<? extends E> poolItemFactory, final int initPoolSize,
            final IAccessManager<? super E> accessManager, final Executor createItemExecutor,
            final IPoolItemRemoveListener<? super E> poolItemRemoveCB, final Executor removeItemExecutor) {
        if (initPoolSize < 0 || poolItemFactory == null || accessManager == null) {
            throw new IllegalArgumentException();
        }
        factoryState = FactoryState.WORKING;
        lock = LockFactory.createLock();
        signalClearableCondition = SignalClearableConditionProxy.createProxy(lock.newCondition());
        sizeMutex = new Object();
        scheduledCreateItemCount = 0;
        delayedCount = 0;
        availablePoolItemList = new ArrayList<E>();
        unmodifiableAvailablePoolItemList = Collections.unmodifiableList(availablePoolItemList);
        this.accessManager = accessManager;
        this.poolItemFactory = poolItemFactory;
        expectedSize = initPoolSize;
        for (int i = 0; i < initPoolSize; i++) {
            E poolItem = null;
            do {
                try {
                    poolItem = poolItemFactory.createPoolItem();
                } catch (RuntimeException e) {
                    LOG.error("Factory failed", e);
                } catch (CreatePoolItemException e) {
                    LOG.debug("Factory failed", e);
                    Long brokenTime = e.getBrokenTime();
                    if (brokenTime == null) {
                        factoryState = FactoryState.BROKEN;
                    } else {
                        factoryState = FactoryState.TEMPORARY_BROKEN;
                        temporaryBrokenFactoryEndTime = System.currentTimeMillis() + brokenTime;
                    }
                    brokenScheduledCreateItemCount = initPoolSize - i;
                    break;
                }
            } while (poolItem == null);
            availablePoolItemList.add(poolItem);
        }
        usedPoolItem = 0;
        ActionSchedulerFactory actionSchedulerFactory = new ActionSchedulerFactory();
        createItemActionScheduler = actionSchedulerFactory.createActionScheduler(new IAction() {

            public Boolean call() {
                return createNewItem();
            }

        }, createItemExecutor);
        this.poolItemRemoveCB = poolItemRemoveCB;
        removeItemActionScheduler = actionSchedulerFactory.createActionScheduler(new IAction() {

            public Boolean call() {
                return callNextRemoveCallBack();
            }

        }, removeItemExecutor);
    }

    protected int getOneItem() {
        int position = 0;
        try {
            position = accessManager.choosePoolItemToGet(unmodifiableAvailablePoolItemList);
            if (position != IAccessManager.DELAY_GET && (position < 0 || position >= availablePoolItemList.size())) {
                LOG.error("Access manager returns {0} which is an invalid value", position);
                position = 0;
            }
        } catch (RuntimeException e) {
            LOG.error("Access manager exception", e);
        }
        return position;
    }

    protected void createOneItem(final E poolItem) {
        int position = 0;
        try {
            position = accessManager.createPoolItem(unmodifiableAvailablePoolItemList, poolItem);
            if (position < 0 || position > availablePoolItemList.size()) {
                LOG.error("Access manager returns {0} which is an invalid value", position);
                position = 0;
            }
        } catch (RuntimeException e) {
            LOG.error("Access manager exception", e);
        }
        availablePoolItemList.add(position, poolItem);
    }

    protected void putOneItem(final E poolItem) {
        int position = 0;
        try {
            position = accessManager.putPoolItem(unmodifiableAvailablePoolItemList, poolItem);
            if (position < 0 || position > availablePoolItemList.size()) {
                LOG.error("Access manager returns {0} which is an invalid value", position);
                position = 0;
            }
        } catch (RuntimeException e) {
            LOG.error("Access manager exception", e);
        }
        availablePoolItemList.add(position, poolItem);
    }

    protected void releaseOneItem() {
        int position = 0;
        try {
            position = accessManager.choosePoolItemToRelease(unmodifiableAvailablePoolItemList);
            if (position < 0 || position >= availablePoolItemList.size()) {
                LOG.error("Access manager returns {0} which is an invalid value", position);
                position = 0;
            }
        } catch (RuntimeException e) {
            LOG.error("Access manager exception", e);
        }
        releasePoolItem(availablePoolItemList.remove(position));
    }

    protected void removeOneItem(final E poolItem) {
        try {
            accessManager.removePoolItem(poolItem);
        } catch (RuntimeException e) {
            LOG.error("Access manager exception", e);
        }
        releasePoolItem(poolItem);
    }


    /**
     * @param expectedSize the expectedSize to set
     */
    public void setExpectedSize(final int expectedSize) {
        if (expectedSize < 0) {
            throw new IllegalArgumentException();
        }
        LOG.debug("Want mutex for setting size to {0}", expectedSize);
        int instanceToCreate = 0;
        int instanceToDelete = 0;

        synchronized (sizeMutex) {
            lock.lock();
            try {
                if (!sizeLockedToZero) {
                    if (this.expectedSize != expectedSize) {
                        LOG.debug("size will change from {0} to {1}", this.expectedSize, expectedSize);
                        this.expectedSize = expectedSize;
                    }
                } else {
                    if (this.expectedSize != 0) {
                        LOG.debug("size will change from {0} to zero", this.expectedSize);
                        this.expectedSize = 0;
                    }
                }
                // calcul instance to create (or delete)
                int unUsedPoolItem = availablePoolItemList.size();
                int total = usedPoolItem + unUsedPoolItem;
                instanceToCreate = this.expectedSize - total;
                if (instanceToCreate < 0) {
                    // we have to many instances (but we cannot delete busy
                    // instances)
                    instanceToDelete = Math.min(-instanceToCreate, unUsedPoolItem);
                    for (int i = 0; i < instanceToDelete; i++) {
                        releaseOneItem();
                    }
                }

                // we have one less instance to create per factory broken
                // if factory broken and no item available signal waiters (that why we need lock)
                switch (factoryState) {
                case BROKEN:
                    instanceToCreate -= brokenScheduledCreateItemCount;
                    if (availablePoolItemList.size() == 0) {
                        signalClearableCondition.signalAllUnclearable();
                    }
                    break;
                case TEMPORARY_BROKEN:
                    if (System.currentTimeMillis() < temporaryBrokenFactoryEndTime) {
                        instanceToCreate -= brokenScheduledCreateItemCount;
                        if (availablePoolItemList.size() == 0) {
                            signalClearableCondition.signalAllUnclearable();
                        }
                        break;
                    }
                    if (brokenScheduledCreateItemCount != 0) {
                        instanceToCreate += brokenScheduledCreateItemCount;
                        brokenScheduledCreateItemCount = 0;
                        factoryState = FactoryState.WORKING;
                    }
                    break;
                case WORKING:
                    break;
                default:
                    throw new PoolError();
                }
            } finally {
                lock.unlock();
            }

            // we have one less instance to create per factory working
            instanceToCreate -= scheduledCreateItemCount;

            if (instanceToCreate <= 0) {
                if (brokenScheduledCreateItemCount > 0) {
                    // remove broken schedule imply no call to remove CB
                    brokenScheduledCreateItemCount += instanceToCreate;
                    if (brokenScheduledCreateItemCount < 0) {
                        instanceToCreate = brokenScheduledCreateItemCount;
                        brokenScheduledCreateItemCount = 0;
                    } else {
                        instanceToCreate = 0;
                    }
                }
                // try to cancel useless creation
                int cancelled = createItemActionScheduler.cancelAction(-instanceToCreate);
                // remove canceled from scheduled count
                scheduledCreateItemCount -= cancelled;
            } else {
                switch (factoryState) {
                case BROKEN:
                case TEMPORARY_BROKEN:
                    brokenScheduledCreateItemCount += instanceToCreate;
                    instanceToCreate = 0;
                    break;
                case WORKING:
                    scheduledCreateItemCount += instanceToCreate;
                    break;
                default:
                    throw new Error();
                }
            }
        }
        if (instanceToDelete > 0 && poolItemRemoveCB != null) {
            removeItemActionScheduler.scheduleAction(instanceToDelete);
        }
        if (instanceToCreate > 0) {
            // ask to call 'instanceToCreate' times the createNewItem() method
            createItemActionScheduler.scheduleAction(instanceToCreate);
        }
    }

    public void lockSizeToZero() {
        synchronized (sizeMutex) {
            sizeLockedToZero = true;
        }
    }

    public void unlockSizeToZero() {
        synchronized (sizeMutex) {
            sizeLockedToZero = false;
        }
    }


    /**
     * @return true if the call to poolItemFactory was done false otherwise
     */
    protected boolean createNewItem() {
        int instanceToCreate = 0;

        E poolItem = null;
        CreatePoolItemException createPoolItemException = null;
        do {
            synchronized (sizeMutex) {
                if (createPoolItemException != null) {
                    lock.lock();
                    try {
                        Long brokenTime = createPoolItemException.getBrokenTime();
                        if (brokenTime == null) {
                            factoryState = FactoryState.BROKEN;
                            if (availablePoolItemList.size() == 0) {
                                signalClearableCondition.signalAllUnclearable();
                            }
                        } else {
                            switch (factoryState) {
                            case BROKEN:
                                break;
                            case TEMPORARY_BROKEN:
                                temporaryBrokenFactoryEndTime = Math.max(temporaryBrokenFactoryEndTime, System
                                        .currentTimeMillis()
                                        + brokenTime);
                                break;
                            case WORKING:
                                temporaryBrokenFactoryEndTime = System.currentTimeMillis() + brokenTime;
                                factoryState = FactoryState.TEMPORARY_BROKEN;
                                if (availablePoolItemList.size() == 0) {
                                    signalClearableCondition.signalAllUnclearable();
                                }
                                break;
                            default:
                                throw new Error();
                            }
                        }
                    } finally {
                        lock.unlock();
                    }
                }
                switch (factoryState) {
                case BROKEN:
                    scheduledCreateItemCount--;
                    brokenScheduledCreateItemCount++;
                    return false;
                case TEMPORARY_BROKEN:
                    scheduledCreateItemCount--;
                    brokenScheduledCreateItemCount++;
                    return false;
                case WORKING:
                    break;
                default:
                    throw new Error();
                }

                lock.lock();
                try {
                    // calcul instance to create (or delete)
                    int unUsedPoolItem = availablePoolItemList.size();
                    int total = usedPoolItem + unUsedPoolItem;
                    instanceToCreate = expectedSize - total;
                    if (instanceToCreate <= 0) {
                        // we have to many instances (however we should manage
                        // remove)
                        if (instanceToCreate != 0 && unUsedPoolItem != 0) {
                            LOG.error("Issue in algorithm : check remove / put methods");
                        }
                        // creation is useless (cancel operation was called too
                        // late)
                        scheduledCreateItemCount--;
                        if (scheduledCreateItemCount == 0) {
                            sizeMutex.notifyAll();
                        }
                        return false;
                    }
                } finally {
                    lock.unlock();
                }
            }

            LOG.debug("create new item");
            try {
                poolItem = poolItemFactory.createPoolItem();
            } catch (RuntimeException e) {
                LOG.error("Factory failed", e);
            } catch (CreatePoolItemException e) {
                LOG.debug("Factory failed", e);
                createPoolItemException = e;
            }

        } while (poolItem == null);

        boolean instanceDeleted = false;

        synchronized (sizeMutex) {

            lock.lock();
            try {
                createOneItem(poolItem);
                // calcul instance to create (or delete)
                int unUsedPoolItem = availablePoolItemList.size();
                int total = usedPoolItem + unUsedPoolItem;
                instanceToCreate = this.expectedSize - total;
                if (instanceToCreate < 0) {
                    // we just create an useless instance
                    int instanceToDelete = Math.min(-instanceToCreate, unUsedPoolItem);
                    if (instanceToDelete != 1) {
                        LOG.error("Issue in algorithm : check remove / put methods");
                    }
                    releaseOneItem();
                    instanceDeleted = true;
                } else {
                    // our new instance can be use
                    signalClearableCondition.signal();
                }
                if (delayedCount != 0) {
                    // at least a thread wait while some item are available
                    // the reason of that is the DELAY_GET, we have to signalAll
                    // thread in this case
                    // however thanks to lock implementation (clearAllSignal)
                    // all thread will not necessary be wake up.
                    signalClearableCondition.signalAll();
                }

            } finally {
                lock.unlock();
            }
            scheduledCreateItemCount--;
            if (scheduledCreateItemCount == 0) {
                sizeMutex.notifyAll();
            }
        }
        if (instanceDeleted) {
            if (poolItemRemoveCB != null) {
                callNextRemoveCallBack();
            }
        }
        return true;
    }

    /**
     * This method block is no instance are available.
     * @param timeout after this timeout a exception is thrown
     * @return a pool item
     * @throws TimeoutPoolException if timeout
     */
    public E get(final IWaitControl timeout) throws TimeoutPoolException, IllegalTimeoutException, WaiterInterruptedException,
            InterruptedException, PoolFactoryBroken {
        E poolItem = null;
        boolean delayed = false;
        lock.lock();
        try {
            int size = availablePoolItemList.size();
            int poolItemToGetChoosed = IAccessManager.DELAY_GET;
            if (size != 0) {
                poolItemToGetChoosed = getOneItem();
                if (poolItemToGetChoosed == IAccessManager.DELAY_GET) {
                    delayedCount++;
                    delayed = true;
                }
            } else {
                switch (factoryState) {
                case BROKEN:
                    throw new PoolFactoryBroken(null);
                case TEMPORARY_BROKEN:
                    throw new PoolFactoryBroken(temporaryBrokenFactoryEndTime);
                case WORKING:
                    break;
                default:
                    throw new Error();
                }
            }
            if (poolItemToGetChoosed == IAccessManager.DELAY_GET) {
                if (!timeout.canContinueToWait()) {
                    throw new TimeoutPoolException();
                }
                while (poolItemToGetChoosed == IAccessManager.DELAY_GET) {
                    if (!timeout.waitOnCondition(signalClearableCondition)) {
                        throw new TimeoutPoolException();
                    }
                    LOG.debug("wait notified or expired");
                    size = availablePoolItemList.size();
                    if (size != 0) {
                        poolItemToGetChoosed = getOneItem();
                        if (!delayed && poolItemToGetChoosed == IAccessManager.DELAY_GET) {
                            delayedCount++;
                            delayed = true;
                        }
                    } else {
                        if (delayed) {
                            // presume not delayed if no item available
                            delayedCount--;
                            delayed = false;
                        }
                        switch (factoryState) {
                        case BROKEN:
                            throw new PoolFactoryBroken(null);
                        case TEMPORARY_BROKEN:
                            throw new PoolFactoryBroken(temporaryBrokenFactoryEndTime);
                        case WORKING:
                            break;
                        default:
                            throw new Error();
                        }
                    }
                }
            }
            usedPoolItem++;
            poolItem = availablePoolItemList.remove(poolItemToGetChoosed);
            if (size == 1) {
                // no more item to distribute so we avoid wake up of other
                // threads.
                signalClearableCondition.clearAllSignal();
            }
            if (LOG.isDebugEnabled()) {
                LOG.debug("pool {0}/{1}", usedPoolItem, availablePoolItemList.size() + usedPoolItem);
            }
        } finally {
            if (delayed) {
                delayedCount--;
            }
            lock.unlock();
        }
        return poolItem;
    }

    /**
     * Because basic pool do not delegate, we can use one instance.
     */
    protected static final InfiniteWaitControl INFINITE_WAIT_CONTROL = new InfiniteWaitControl();

    /**
     * @return a non null instance
     * @throws WaiterInterruptedException if error occurs
     */
    public E get() throws WaiterInterruptedException, InterruptedException, PoolFactoryBroken {
        try {
            return get(INFINITE_WAIT_CONTROL);
        } catch (TimeoutPoolException e) {
            throw new Error("get(INFINITE_TIMEOUT) should not return a TimeoutPoolException");
        } catch (IllegalTimeoutException e) {
            throw new Error("get(INFINITE_TIMEOUT) should not return a IllegalTimeoutException");
        }
    }

    /**
     * This method will never block. If the pool owns too many instances,
     * release method will be apply on pool item.
     * @param poolItem a pool item
     * @throws NotABusyPoolItemException if error occurs
     */
    public void put(final E poolItem) throws NotABusyPoolItemException {
        if (poolItem == null) {
            throw new NotABusyPoolItemException();
        }
        boolean instanceDeleted = false;
        lock.lock();
        try {
            if (usedPoolItem <= 0) {
                throw new NoBusyPoolItemException();
            }
            int remainToDelete = usedPoolItem - expectedSize;
            usedPoolItem--;
            putOneItem(poolItem);
            if (remainToDelete <= 0) {
                // add to available if no remain to delete
                if (delayedCount == 0) {
                    signalClearableCondition.signal();
                } else {
                    // at least a thread wait while some item are available
                    // the reason of that is the DELAY_GET, we have to signalAll
                    // thread in this case
                    // however thanks to lock implementation (clearAllSignal)
                    // all thread will not necessary be wake up.
                    signalClearableCondition.signalAll();
                }
            } else {
                // let gc run on poolItem otherwise
                releaseOneItem();
                instanceDeleted = true;
            }
            if (LOG.isDebugEnabled()) {
                LOG.debug("pool {0}/{1} (used/total)", usedPoolItem, availablePoolItemList.size() + usedPoolItem);
            }
        } finally {
            lock.unlock();
        }
        if (instanceDeleted) {
            if (poolItemRemoveCB != null) {
                removeItemActionScheduler.scheduleAction(1);
            }
        }
    }

    /**
     * @return the poolItemFactory
     */
    protected IPoolItemFactory<? extends E> getPoolItemFactory() {
        return poolItemFactory;
    }

    /**
     * @return the releaseListener
     */
    protected IAccessManager<? super E> getReleaseListener() {
        return accessManager;
    }

    /**
     * @param usedPoolItem the usedPoolItem to set
     */
    protected void setUsedPoolItem(final int usedPoolItem) {
        this.usedPoolItem = usedPoolItem;
    }

    /**
     * @return the mutex to sync before calling protected methods
     */
    protected List<E> getAvailablePoolItemList() {
        return availablePoolItemList;
    }

    /**
     * @return the expectedSize
     */
    protected int getExpectedSize() {
        return expectedSize;
    }

    /**
     * @return the usedPoolItem
     */
    protected int getUsedPoolItem() {
        return usedPoolItem;
    }

    /**
     * @return the unUsedPoolItem
     */
    protected int getUnUsedPoolItem() {
        return availablePoolItemList.size();
    }

    /**
     * @return the lock
     */
    protected final Lock getLock() {
        return lock;
    }

    /**
     * @return the signalClearableCondition
     */
    protected ISignalClearableCondition getSignalClearableCondition() {
        return signalClearableCondition;
    }

    /**
     * @return the delayedCount
     */
    protected int getDelayedCount() {
        return delayedCount;
    }

    /**
     * @param delayedCount the delayedCount to set
     */
    protected void setDelayedCount(final int delayedCount) {
        this.delayedCount = delayedCount;
    }

    public void signalAllWaiters() {
        lock.lock();
        try {
            if (LOG.isDebugEnabled()) {
                LOG.debug("begin interrupt {0}/{1} (used/total)", usedPoolItem, availablePoolItemList.size() + usedPoolItem);
            }
            signalClearableCondition.signalAllUnclearable();
        } finally {
            lock.unlock();
        }
    }

    // may be complete
    private RemoveEntry<E> firstTreatedRemoveEntry;

    // not treated
    private RemoveEntry<E> headRemoveEntry;

    // latest not treated
    private RemoveEntry<E> queueRemoveEntry;

    private Object removeMutex = new Object();

    protected void releasePoolItem(final E releasePoolItem) {
        if (poolItemRemoveCB != null) {
            synchronized (removeMutex) {
                RemoveEntry<E> nextQueueRemoveEntry = new RemoveEntry<E>(releasePoolItem, null);
                if (headRemoveEntry == null) {
                    headRemoveEntry = nextQueueRemoveEntry;
                }
                if (queueRemoveEntry == null) {
                    firstTreatedRemoveEntry = nextQueueRemoveEntry;
                } else {
                    queueRemoveEntry.setNext(nextQueueRemoveEntry);
                }
                queueRemoveEntry = nextQueueRemoveEntry;
            }
        }
    }

    public void clearPool() throws InterruptedException {
        setExpectedSize(0);
        synchronized (sizeMutex) {
            while (scheduledCreateItemCount != 0) {
                sizeMutex.wait();
                if (scheduledCreateItemCount < 0) {
                    throw new Error("wait not possible, pool buggy");
                }
            }
        }
    }

    public void waitAllRemoveListenerCompleted() throws InterruptedException {
        if (poolItemRemoveCB == null) {
            return;
        }
        RemoveEntry<E> firstWaitForRemoveEntry;
        RemoveEntry<E> lastWaitForRemoveEntry;
        synchronized (removeMutex) {
            firstWaitForRemoveEntry = firstTreatedRemoveEntry;
            if (firstWaitForRemoveEntry == null) {
                return;
            }
            lastWaitForRemoveEntry = queueRemoveEntry;
        }

        synchronized (lastWaitForRemoveEntry) {
            while (!lastWaitForRemoveEntry.isCallBackComplete()) {
                lastWaitForRemoveEntry.wait();
            }
        }

        // all after callback before (should be terminated)
        RemoveEntry<E> currentWaitForRemoveEntry = firstWaitForRemoveEntry;
        while (currentWaitForRemoveEntry != lastWaitForRemoveEntry) {
            synchronized (currentWaitForRemoveEntry) {
                while (!currentWaitForRemoveEntry.isCallBackComplete()) {
                    currentWaitForRemoveEntry.wait();
                }
            }
            // next exists (and cannot be changed) because this is not the last
            // element [lastWaitForRemoveEntry is before (or is) the last
            // element]
            currentWaitForRemoveEntry = currentWaitForRemoveEntry.getNext();
        }
    }

    protected boolean callNextRemoveCallBack() {
        RemoveEntry<E> previousHeadRemoveEntry;
        synchronized (removeMutex) {
            previousHeadRemoveEntry = headRemoveEntry;
            if (previousHeadRemoveEntry == null) {
                return false;
            }
            headRemoveEntry = headRemoveEntry.getNext();
        }


        try {
            poolItemRemoveCB.poolItemRemoved(previousHeadRemoveEntry.getPoolItem());
        } catch (RuntimeException e) {
            LOG.error("poolItemRemoved failed", e);
        }

        synchronized (previousHeadRemoveEntry) {
            previousHeadRemoveEntry.setCallBackComplete(true);
            previousHeadRemoveEntry.notifyAll();
        }
        synchronized (removeMutex) {
            if (firstTreatedRemoveEntry == previousHeadRemoveEntry) {
                firstTreatedRemoveEntry = firstTreatedRemoveEntry.getNext();
                while (firstTreatedRemoveEntry != null) {
                    synchronized (firstTreatedRemoveEntry) {
                        if (firstTreatedRemoveEntry.isCallBackComplete()) {
                            firstTreatedRemoveEntry = firstTreatedRemoveEntry.getNext();
                        } else {
                            break;
                        }
                    }
                }
                if (firstTreatedRemoveEntry == null) {
                    queueRemoveEntry = null;
                }
            }
        }
        return true;
    }

    public void remove(final E poolItem) throws NotABusyPoolItemException {
        if (poolItem == null) {
            throw new NotABusyPoolItemException();
        }
        boolean scheduleCreation = false;

        synchronized (sizeMutex) {
            int instanceToCreate = 0;

            lock.lock();
            try {
                if (usedPoolItem <= 0) {
                    throw new NoBusyPoolItemException();
                }
                removeOneItem(poolItem);
                usedPoolItem--;
                // calcul instance to create (or delete)
                int unUsedPoolItem = availablePoolItemList.size();
                int total = usedPoolItem + unUsedPoolItem;
                instanceToCreate = expectedSize - total;
                if (instanceToCreate < 0 && unUsedPoolItem != 0) {
                    LOG.error("issue in setExpectedSize, remove method should not make this stuff");
                }
                if (LOG.isDebugEnabled()) {
                    LOG.debug("pool {0}/{1} (used/total)", usedPoolItem, availablePoolItemList.size() + usedPoolItem);
                }
            } finally {
                lock.unlock();
            }

            // we have one less instance to create per factory working
            instanceToCreate -= scheduledCreateItemCount;
            if (instanceToCreate > 0) {
                scheduledCreateItemCount++;
                scheduleCreation = true;
                if (instanceToCreate != 1) {
                    LOG.error("issue in setExpectedSize, remove method should not schedule more than one instance creation");
                }
            }
        }
        if (poolItemRemoveCB != null) {
            removeItemActionScheduler.scheduleAction(1);
        }
        if (scheduleCreation) {
            createItemActionScheduler.scheduleAction(1);
        }
    }

    /**
     * @return the unmodifiableAvailablePoolItemList
     */
    protected List<E> getUnmodifiableAvailablePoolItemList() {
        return unmodifiableAvailablePoolItemList;
    }

    protected long getTemporaryBrokenFactoryEndTime() {
        return temporaryBrokenFactoryEndTime;
    }

    protected FactoryState getFactoryState() {
        return factoryState;
    }

}
